import copy
import json
import subprocess
import torch

import logging
LOG = logging.getLogger(__name__)


def load_json(jsonfile):
    with open(jsonfile, "r") as jf:
        res = json.load(jf)
    return res


def save_config(config, filepath):
    with open(filepath, "w") as outfile:    
        json.dump(config, outfile, indent=4)


def apply_diff(base_cfg, diff):
    cfg = copy.deepcopy(base_cfg)
    for k, v in diff.items():
        if isinstance(v, dict):
            cfg[k] = apply_diff(base_cfg[k], v)
        else:
            cfg[k] = v
    return cfg


def execute_bash_command(bash_command_string):
    bash_return = subprocess.run(bash_command_string.split(), timeout=20)
    return bash_return


def get_torch_device(force_cpu=False, use_mps_if_available=False):
    try:
        if torch.cuda.is_available() and not force_cpu:
            device_string = "cuda"
        elif use_mps_if_available and torch.backends.mps.is_available() and not force_cpu:
            device_string = "mps"
        else:
            device_string = "cpu"
    except Exception as exc:
        LOG.error(f'Exception: {exc}')
        device_string = "cpu"
    LOG.info(f'Using device: {device_string}')
    device = torch.device(device_string)
    return device



import abc
import tempfile
from typing import Dict

import MDAnalysis as mda

from src.chainsaw.run_chainsaw import predict_chopping_chainsaw_from_pdb
from src.merizo.run_merizo import predict_chopping_merizo_from_pdb


class Cluster(abc.ABC):
    @abc.abstractmethod
    def predict_from_pdb(self, pdb_path: str, model_params: Dict = None):
        """
        Predicts the chopping of a protein from a PDB file
        """
        raise NotImplementedError

    def predict_from_universe(self, universe: mda.Universe, model_params: Dict = None):
        """
        Predicts the chopping of a protein from a MDAnalysis Universe
        """
        with tempfile.TemporaryFile() as temp_pdb:
            universe.atoms.write(temp_pdb, format='pdb')
            return self.predict_from_pdb(temp_pdb.name, model_params)


class MerizoCluster(Cluster):
    def predict_from_pdb(self, pdb_path: str, model_params: Dict = None):
        """
        Predicts the chopping of a protein from a PDB file
        """
        return predict_chopping_merizo_from_pdb(pdb_path, **(model_params or {}))


class ChainsawCluster(Cluster):
    def predict_from_pdb(self, pdb_path: str, model_params: Dict = None):
        """
        Predicts the chopping of a protein from a PDB file
        """
        return predict_chopping_chainsaw_from_pdb(pdb_path, **(model_params or {}))



import torch
from math import log2, floor
from torch import nn
from einops import rearrange

# Symmetric ALiBi relative positional bias
class AlibiPositionalBias(nn.Module):
    def __init__(self, heads, slope_factor):
        super().__init__()
        self.heads = heads
        slopes = torch.Tensor(self._get_slopes(heads))
        slopes = rearrange(slopes, 'h -> () h () ()')
        self.register_buffer('slopes', slopes, persistent=False)
        self.register_buffer('bias', None, persistent=False)
        
        self.slope_factor = slope_factor

    @staticmethod
    def _get_slopes(heads):
        def get_slopes_power_of_2(n):
            start = (2**(-2**-(log2(n)-3)))
            ratio = start
            return [start*ratio**i for i in range(n)]

        if log2(heads).is_integer():
            return get_slopes_power_of_2(heads)

        closest_power_of_2 = 2 ** floor(log2(heads))
        return get_slopes_power_of_2(closest_power_of_2) + get_slopes_power_of_2(2 * closest_power_of_2)[0::2][:heads-closest_power_of_2]

    def forward(self, residx, clip=False):
        relative_position = residx.unsqueeze(0) - residx.unsqueeze(1)
        bias = torch.abs(relative_position)

        # Clip attention bias offsets at 32, along lines of AlphaFold2 relative pos encoding
        bias = bias.clip(max=32) if clip else bias
        bias = bias.unsqueeze(0).expand(1, self.heads, -1, -1)

        return bias * -(self.slopes * self.slope_factor)


"""
Created by Jude Wells 2023-04-20
Objective is to create a secondary structure matrix for each protein
1) renumbers the pdb file (1-indexed)
2) runs stride to get secondary structure file
3) parses the secondary structure file to create a matrix of secondary structure

# should take around 3 minutes for 1000 structures
"""

import os
import re
import subprocess
import sys

import numpy as np

from src.chainsaw.constants import REPO_ROOT
import logging
LOG = logging.getLogger(__name__)


def calculate_ss(pdbfile, chain, stride_path, ssfile='pdb_ss'):
    assert os.path.exists(pdbfile)
    with open(ssfile, 'w') as ssout_file:
        args = [stride_path, pdbfile, '-r' + chain]
        LOG.info(f"Running command: {' '.join(args)}")
        try:
            subprocess.run(args, 
                        stdout=ssout_file, 
                        stderr=subprocess.DEVNULL, 
                        check=True)
        except subprocess.CalledProcessError as e:
            LOG.warning(f"Stride failed on {pdbfile}, creating empty file")
            pass


def make_ss_matrix(ss_path, nres):
    # create matrices for helix and strad residues where entry ij = 1 if i and j are in the same helix or strand
    with open(ss_path) as f:
        lines = f.readlines()
    type_set = set()
    helix  = np.zeros([nres, nres], dtype=np.float32)
    strand = np.zeros([nres, nres], dtype=np.float32)
    for line in lines:
        if line.startswith('LOC'):
            start = int(re.sub('\D', '',line[22:28].strip()))
            end = int(re.sub('\D', '',line[40:46].strip()))
            type = line[5:17].strip()
            type_set.add(type)
            if type in ['AlphaHelix', '310Helix']:
                helix[start-1:end, start-1:end] = 1
            elif type == 'Strand':
                strand[start-1:end, start-1:end] = 1
        elif line.startswith('ASG'):
            break
    return helix, strand


def renum_pdb_file(pdb_path, output_pdb_path):
    pdb_reres_path = REPO_ROOT / 'utils/pdb_reres.py'
    with open(output_pdb_path, "w") as output_file:
        print("Shell command: ", [sys.executable, str(pdb_reres_path), pdb_path])
        subprocess.run([sys.executable, str(pdb_reres_path), pdb_path],
                       stdout=output_file,
                       check=True,
                       text=True)
 

def main(chain_ids, pdb_dir, feature_dir, stride_path, reres_path, savedir, job_index=0):
    os.makedirs(savedir, exist_ok=True)
    os.makedirs(os.path.join(savedir, '2d_features'), exist_ok=True)
    for chain_id in chain_ids:
        try:
            pdb_path = os.path.join(pdb_dir, chain_id + '.pdb')
            if os.path.exists(pdb_path):
                features = np.load(os.path.join(feature_dir, chain_id + '.npz'))['arr_0']
                nres = features.shape[-1]
                LOG.info("Processing", pdb_path)
                chain = chain_id[4]
                output_pdb_path = os.path.join(savedir, f"{job_index}.pdb") # this gets overwritten to save memory
                file_nres = renum_pdb_file(pdb_path, reres_path, output_pdb_path)
                if nres != file_nres:
                    with open(os.path.join(savedir, 'error.txt'), 'a') as f:
                        msg = f' residue number mismatch (from features) {nres}, (from pdb file) {file_nres}'
                        f.write(chain_id + msg + '\n')
                ss_filepath = os.path.join(savedir, f'pdb_ss{job_index}.txt') # this gets overwritten to save memory
                calculate_ss(output_pdb_path, chain, stride_path, ssfile=ss_filepath)
                helix, strand = make_ss_matrix(ss_filepath, nres=nres)
                np.savez_compressed(
                    os.path.join(*[savedir, '2d_features', chain_id + '.npz']), 
                    np.stack((features, helix, strand), axis=0))
        except Exception as e:
            with open(os.path.join(savedir, 'error.txt'), 'a') as f:
                f.write(chain_id + str(e) + '\n')



from __future__ import annotations

import os
from pathlib import Path

import numpy as np
import torch
from torch.types import Device

from src.merizo.model.network import Merizo
from src.merizo.model.utils.features import generate_features_domain
from src.merizo.model.utils.utils import (
    instance_matrix,
    clean_domains,
    clean_singletons,
    get_ids,
    remap_ids,
    shuffle_ids,
    separate_components, get_device,
)
from src.shared.structure import get_model_structure

MIN_DOMAIN_SIZE = 50
MIN_FRAGMENT_SIZE = 10
DOM_AVE = 200
CONF_THRESHOLD = 0.75


def iterative_segmentation(
        network: torch.nn.Module, features: dict, max_iterations: int
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Iteratively segment protein domains with a trained network.
    """
    n_iterations, ignore_index, iterate = 0, [], True
    domain_ids, conf_res = features['domain_ids'], features['conf_res']

    while iterate and n_iterations < max_iterations:
        ids, _ = get_ids(domain_ids)
        # Identify domains that are large enough to be re-segmented
        unique_ids = {
            d.item(): conf_res[domain_ids == d].mean()
            for d in ids
            if d.item() not in ignore_index and (domain_ids == d).sum() > DOM_AVE
        }

        if not unique_ids:
            break

        # Re-segment each domain that meets criteria
        counter = 1
        for old_did, _ in unique_ids.items():
            mask = domain_ids == old_did
            new_ids, new_conf = network(features, mask=mask)
            _, ndoms_ = get_ids(new_ids)

            if ndoms_ == 1:  # If it stayed one domain, ignore future re-segmentation
                ignore_index.append(old_did)
            else:
                # Offset new IDs to avoid collisions
                offset_ids = new_ids + (counter * network.no_classes)
                offset_ids[new_ids == 0] = 0
                domain_ids[mask] = offset_ids
                conf_res[mask] = new_conf
                counter += 1

        n_iterations += 1

    return domain_ids, conf_res


def read_split_weight_files(directory: str) -> dict:
    """
    Read .pt weight files in a directory into a single dictionary.
    """
    weights = {}
    for f in os.listdir(directory):
        if f.endswith('.pt'):
            weights.update(torch.load(os.path.join(directory, f), weights_only=True))

    return weights


def segment(
        pdb_path: str,
        network: torch.nn.Module,
        device: Device,
        length_conditional_iterate: bool,
        iterate: bool,
        max_iterations: int,
        shuffle_indices: bool,
        pdb_chain: str = "A",
) -> dict:
    """
    Segment domains in a protein structure.
    """
    features = generate_features_domain(pdb_path, device, pdb_chain)
    if length_conditional_iterate and features['nres'] > 512:
        iterate = True

    features['domain_ids'], features['conf_res'] = network(features)

    # Optionally iterate multiple times if large
    if iterate and features['nres'] > DOM_AVE * 2:
        features['domain_ids'], features['conf_res'] = iterative_segmentation(
            network, features, max_iterations
        )

    features['domain_map'] = instance_matrix(features['domain_ids'])[0]
    features['domain_ids'] = separate_components(features)

    # Clean up small domains or singletons
    if torch.unique(features['domain_ids']).numel() > 1:
        features['domain_ids'] = clean_domains(features['domain_ids'], MIN_DOMAIN_SIZE)
        features['domain_ids'] = clean_singletons(features['domain_ids'], MIN_FRAGMENT_SIZE)

    features['domain_map'] = instance_matrix(features['domain_ids'])[0]
    features['conf_global'] = features['conf_res'].mean()
    features['ndom'] = get_ids(features['domain_ids'])[1]
    features['domain_ids'] = (
        shuffle_ids(features['domain_ids']) if shuffle_indices else remap_ids(features['domain_ids'])
    )

    return features


def features_to_chopping_string(features) -> str:
    # Convert domain IDs into consecutive-residue group strings
    domain_ids = np.array(features["domain_ids"])
    outputs = []
    for d in np.unique(domain_ids):
        matching = np.where(domain_ids == d)[0]
        # Find consecutive stretches
        groups, curr = [], []
        for idx in matching:
            if not curr or idx == curr[-1] + 1:
                curr.append(idx)
            else:
                groups.append(curr)
                curr = [idx]
        if curr:
            groups.append(curr)
        # Format each group into "start-end" or single residue
        chunked = [f"{g[0] + 1}-{g[-1] + 1}" if len(g) > 1 else f"{g[0] + 1}" for g in groups]
        outputs.append("_".join(chunked))
    output = ",".join(outputs)
    return output


def predict_chopping_merizo_from_pdb(
        input_path: str,
        device: str = 'cpu',
        max_iterations: int = 3,
        length_conditional_iterate: bool = False,
        iterate: bool = False,
        shuffle_indices: bool = False,
        pdb_chain: str = None,
) -> str:
    """
    Run Merizo on a PDB file and return the chopping string.
    """
    device = get_device(device)
    network = Merizo().to(device)

    weights_dir = Path(__file__).parent.parent / "models" / "merizo"
    network.load_state_dict(read_split_weight_files(str(weights_dir)), strict=True)
    network.eval()

    if not os.path.exists(input_path):
        raise FileNotFoundError(f"{input_path} does not exist.")

    with torch.no_grad():
        _, _ = os.path.splitext(input_path)

        # If no chain is provided, pick the first chain from the structure
        if not pdb_chain:
            structure = get_model_structure(input_path)
            chains = [c.id for c in structure.get_chains()]
            if not chains:
                raise ValueError(f"No chains found in {input_path}")
            pdb_chain = chains[0]

        features = segment(
            pdb_path=input_path,
            network=network,
            device=device,
            length_conditional_iterate=length_conditional_iterate,
            iterate=iterate,
            max_iterations=max_iterations,
            shuffle_indices=shuffle_indices,
            pdb_chain=pdb_chain,
        )

        return features_to_chopping_string(features)



import numpy as np
import torch
from src.chainsaw.domain_assignment.sparse_lowrank import greedy_V


class BaseAssigner:

    def __call__(self, y_pred, uncertainty=False):
        assert y_pred.ndim == 2
        if torch.is_tensor(y_pred):
            y_pred = y_pred.detach().cpu().numpy()
        return self.assign_domains(y_pred)


class SparseLowRank(BaseAssigner):

    """Use Brooks' method to generate a matrix of cluster assignments,
    and post-process to remove any excessively small clusters.
    """

    def __init__(self, N_iters=3, K_init=4, linker_threshold=5, cost_type="mse"):
        self.N_iters = N_iters
        self.K_init = K_init
        self.linker_threshold = linker_threshold
        self.cost_type = cost_type

    def get_entropy(self, y_pred):
        """calculate the entropy of the upper triangle of the matrix
        used for uncertainty calculation but does not correlate with NDO as well
        as likelihood of assignment under y_pred
        """
        return np.triu(-1 * y_pred * np.log(y_pred) - (1 - y_pred) * np.log(1 - y_pred), 1).sum()


    def assign_domains(self, y_pred):
        # N x K, columns are then indicator vectors
        epsilon = 1e-6
        y_pred = np.clip(y_pred, epsilon, 1-epsilon)
        V, loss = greedy_V(y_pred, N_iters=self.N_iters, K_init=self.K_init, cost_type=self.cost_type)
        K = V.shape[-1]
        A = V@V.T
        average_likelihood = np.exp((A * np.log(y_pred) + (1-(A))*np.log(1-y_pred)).mean())
        # throw away small clusters
        V = np.where(
            V.sum(0, keepdims=True) < self.linker_threshold,  # 1, K
            np.zeros_like(V),
            V,
        )

        assignments = {
            "linker": np.argwhere((V == 0).sum(-1) == K).reshape(-1)
        }

        domain_ix = 1
        for col_ix in range(K):
            cluster_inds = np.argwhere(V[:, col_ix] == 1).reshape(-1)
            if cluster_inds.size > 0:
                assignments[f"domain_{domain_ix}"] = cluster_inds
                domain_ix += 1
        return assignments, round(average_likelihood, 4)


class SpectralClustering(BaseAssigner):
    """Challenge is how to determine n_clusters.
    
    We can look at eigengaps
    Or use MSE, running clustering multiple times.
    Or use Brooks' method and then refine with SC.
    """

    def __init__(self, n_cluster_method="mse", max_domains=10):
        self.n_cluster_method = n_cluster_method
        self.max_domains = max_domains  # TODO - make this flexible or length dependent.

    def assign_domains(self, y_pred):
        raise NotImplementedError()



import torch
import torch.nn as nn

from .nndef_ipa_primitives import *
from rotary_embedding_torch import RotaryEmbedding

class StructureModuleTransition(nn.Module):
    def __init__(self, c, num_layers, dropout_rate):
        super(StructureModuleTransition, self).__init__()

        self.layers = nn.ModuleList()
        for _ in range(num_layers):
            self.layers.append(
                nn.GRU(
                    input_size=c,
                    hidden_size=c // 2,
                    num_layers=2,
                    batch_first=True,
                    dropout=dropout_rate,
                    bidirectional=True
                )
            )

        self.dropout = nn.Dropout(dropout_rate)
        self.layer_norm = LayerNorm(c)

    def forward(self, s):
        for l in self.layers:
            s = l(s)[0]

        s = self.dropout(s)
        s = self.layer_norm(s)

        return s


class InvariantPointAttention(nn.Module):
    """
    Implements Algorithm 22.
    """

    def __init__(
        self,
        c_s: int,
        c_z: int,
        c_hidden: int,
        no_heads: int,
        no_qk_points: int,
        no_v_points: int,
        inf: float = 1e5,
        eps: float = 1e-8,
    ):
        """
        Args:
            c_s:
                Single representation channel dimension
            c_z:
                Pair representation channel dimension
            c_hidden:
                Hidden channel dimension
            no_heads:
                Number of attention heads
            no_qk_points:
                Number of query/key points to generate
            no_v_points:
                Number of value points to generate
        """
        super(InvariantPointAttention, self).__init__()

        self.c_s = c_s
        self.c_z = c_z
        self.c_hidden = c_hidden
        self.no_heads = no_heads
        self.no_qk_points = no_qk_points
        self.no_v_points = no_v_points
        self.inf = inf
        self.eps = eps

        # These linear layers differ from their specifications in the
        # supplement. There, they lack bias and use Glorot initialization.
        # Here as in the official source, they have bias and use the default
        # Lecun initialization.
        hc = self.c_hidden * self.no_heads
        self.linear_q = Linear(self.c_s, hc)
        self.linear_kv = Linear(self.c_s, 2 * hc)

        hpq = self.no_heads * self.no_qk_points * 3
        self.linear_q_points = Linear(self.c_s, hpq)

        hpkv = self.no_heads * (self.no_qk_points + self.no_v_points) * 3
        self.linear_kv_points = Linear(self.c_s, hpkv)

        self.head_weights = nn.Parameter(torch.zeros((no_heads)))
        ipa_point_weights_init_(self.head_weights)

        self.linear_b = Linear(self.c_z, self.no_heads)
        
        self.pair_out = Linear(self.no_heads * self.c_z,
                            self.c_s, init="final")
            
        self.hidden_out = Linear(
            self.no_heads * self.c_hidden, self.c_s, init="final")
        self.points_out = Linear(
            self.no_heads * self.no_v_points * 3, self.c_s, init="final")
        self.points_norm_out = Linear(
            self.no_heads * self.no_v_points, self.c_s, init="final")

        self.softmax = nn.Softmax(dim=-1)
        self.softplus = nn.Softplus()
        
        self.rotary_emb = RotaryEmbedding(dim=self.c_hidden // 2)

    def forward(
        self,
        s: torch.Tensor,
        z: torch.Tensor,
        r: Rigid,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        """
        Args:
            s:
                [*, N_res, C_s] single representation
            z:
                [*, N_res, N_res, C_z] pair representation
            r:
                [*, N_res] transformation object
            mask:
                [*, N_res] mask
        Returns:
            [*, N_res, C_s] single representation update
        """
        
        #######################################
        # Generate scalar and point activations
        #######################################
        # [*, N_res, H * C_hidden]
        q = self.linear_q(s)
        kv = self.linear_kv(s)

        # [*, N_res, H, C_hidden]
        q = q.view(q.shape[:-1] + (self.no_heads, -1))

        # [*, N_res, H, 2 * C_hidden]
        kv = kv.view(kv.shape[:-1] + (self.no_heads, -1))

        # [*, N_res, H, C_hidden]
        k, v = torch.split(kv, self.c_hidden, dim=-1)

        # [*, N_res, H * P_q * 3]
        q_pts = self.linear_q_points(s)

        # This is kind of clunky, but it's how the original does it
        # [*, N_res, H * P_q, 3]
        q_pts = torch.split(q_pts, q_pts.shape[-1] // 3, dim=-1)
        q_pts = torch.stack(q_pts, dim=-1)
        q_pts = r[..., None].apply(q_pts)

        # [*, N_res, H, P_q, 3]
        q_pts = q_pts.view(
            q_pts.shape[:-2] + (self.no_heads, self.no_qk_points, 3)
        )

        # [*, N_res, H * (P_q + P_v) * 3]
        kv_pts = self.linear_kv_points(s)

        # [*, N_res, H * (P_q + P_v), 3]
        kv_pts = torch.split(kv_pts, kv_pts.shape[-1] // 3, dim=-1)
        kv_pts = torch.stack(kv_pts, dim=-1)
        kv_pts = r[..., None].apply(kv_pts)

        # [*, N_res, H, (P_q + P_v), 3]
        kv_pts = kv_pts.view(kv_pts.shape[:-2] + (self.no_heads, -1, 3))

        # [*, N_res, H, P_q/P_v, 3]
        k_pts, v_pts = torch.split(
            kv_pts, [self.no_qk_points, self.no_v_points], dim=-2
        )

        ##########################
        # Compute attention scores
        ##########################

        q = self.rotary_emb.rotate_queries_or_keys(q)
        k = self.rotary_emb.rotate_queries_or_keys(k)

        # [*, H, N_res, N_res]
        a = torch.matmul(
            permute_final_dims(q, (1, 0, 2)),  # [*, H, N_res, C_hidden]
            permute_final_dims(k, (1, 2, 0)),  # [*, H, C_hidden, N_res]
        )
        
        a *= math.sqrt(1.0 / (3 * self.c_hidden))
            
        # [*, N_res, N_res, H]
        b = self.linear_b(z)
        a += (math.sqrt(1.0 / 3) * permute_final_dims(b, (2, 0, 1)))

        # [*, N_res, N_res, H, P_q, 3]
        pt_att = q_pts.unsqueeze(-4) - k_pts.unsqueeze(-5)
        pt_att = pt_att ** 2

        # [*, N_res, N_res, H, P_q]
        pt_att = sum(torch.unbind(pt_att, dim=-1))
        head_weights = self.softplus(self.head_weights).view(
            *((1,) * len(pt_att.shape[:-2]) + (-1, 1))
        )
        head_weights = head_weights * math.sqrt(
            1.0 / (3 * (self.no_qk_points * 9.0 / 2))
        )
        pt_att = pt_att * head_weights

        # [*, N_res, N_res, H]
        pt_att = torch.sum(pt_att, dim=-1) * (-0.5)
        # [*, N_res, N_res]
        square_mask = mask.unsqueeze(-1) * mask.unsqueeze(-2)
        square_mask = self.inf * (square_mask - 1)

        # [*, H, N_res, N_res]
        pt_att = permute_final_dims(pt_att, (2, 0, 1))
        a = a + pt_att

        a = a + square_mask.unsqueeze(-3)

        a = self.softmax(a)
        attn_map = a

        ################
        # Compute output
        ################
        # [*, N_res, H, C_hidden]
        o = torch.matmul(
            a, v.transpose(-2, -3).to(dtype=a.dtype)
        ).transpose(-2, -3)

        # [*, N_res, H * C_hidden]
        o = flatten_final_dims(o, 2)

        # As DeepMind explains, this manual matmul ensures that the operation
        # happens in float32.
        # [*, H, 3, N_res, P_v]
        o_pt = torch.sum(
            (
                a[..., None, :, :, None]
                * permute_final_dims(v_pts, (1, 3, 0, 2))[..., None, :, :]
            ),
            dim=-2,
        )

        # [*, N_res, H, P_v, 3]
        o_pt = permute_final_dims(o_pt, (2, 0, 3, 1))
        o_pt = r[..., None, None].invert_apply(o_pt)

        # [*, N_res, H * P_v]
        o_pt_norm = flatten_final_dims(
            torch.sqrt(torch.sum(o_pt ** 2, dim=-1) + self.eps), 2
        )

        # [*, N_res, H * P_v, 3]
        o_pt = o_pt.reshape(*o_pt.shape[:-3], -1, 3)
        
        s_out = self.hidden_out(o)
        
        # [*, N_res, H, C_z]
        o_pair = torch.matmul(a.transpose(-2, -3), z.to(dtype=a.dtype))

        # [*, N_res, H * C_z]
        o_pair = flatten_final_dims(o_pair, 2)
        
        s_out += self.pair_out(o_pair)

        s_out += self.points_out(
            torch.cat([*torch.unbind(o_pt, dim=-1)], dim=-1))
        
        s_out += self.points_norm_out(o_pt_norm)
        s = s_out

        return s, attn_map



from __future__ import annotations

from itertools import groupby
from operator import itemgetter

import matplotlib.colors as mc
import networkx as nx
import torch
from natsort import natsorted

from src.merizo.model.utils.constants import PRIMES


def get_device(device: str) -> torch.device:
    """Returns the selected device for inference.

    Args:
        device (str): expects 'cuda' or 'cpu' from argparse.

    Returns:
        torch.device: returns selected torch.device object.
    """

    if torch.cuda.is_available() and device == "cuda":
        device = torch.device("cuda")  # GPU

    elif device == "cpu" or (not torch.cuda.is_available() and device == "cuda"):
        device = torch.device("cpu")  # CPU

    elif device == "mps":
        device = torch.device("mps")  # Apple Sllicon

    else:
        print("Device (-d) needs to be either 'cpu', 'cuda' or 'mps'.")
        exit()

    return device


def get_ids(x: torch.tensor) -> tuple[torch.tensor, int]:
    """ Returns a tuple containing unique non-zero indices and 
        the number of unique non-zero indices.

    Args:
        tensor (torch.tensor):      [N] domain_indices

    Returns:
        tuple[torch.tensor, int]:   Tuple of unique non-zero
                                    domain indices and number of 
                                    such indices
    """

    ids = x[x.nonzero()].unique()
    return ids, len(ids)


def remap_ids(x):
    """
    Remap non-zero values in an array of integers beginning from 1 in order of occurrence.
    ** Don't use torch.unique() as the integers will be sorted
    """
    # Get the unique non-zero values in the array
    unique_vals = []
    for val in x:
        if val != 0 and val not in unique_vals:
            unique_vals.append(val.item())

    # Create a mapping from old values to new values
    mapping = {val: i + 1 for i, val in enumerate(unique_vals)}

    # Use the mapping to remap the non-zero values in the array
    remapped_arr = torch.zeros_like(x)
    for i, val in enumerate(x):
        if val != 0:
            remapped_arr[i] = mapping[val.item()]

    return remapped_arr


def shuffle_ids(domain_ids: torch.tensor) -> torch.tensor:
    """ Remap domain indices to increase contrast between
        domains in PyMOL. 

    Args:
        domain_ids (torch.tensor): [N] domain indices [0, ndoms]

    Returns:
        torch.tensor:              [N] new domain indixes [0, ndoms]
    """

    unique_values, _ = get_ids(domain_ids)
    dummy_ids = torch.randperm(len(unique_values)) + 1

    new_ids = torch.zeros_like(domain_ids)
    for u, d in zip(unique_values, dummy_ids):
        new_ids[domain_ids == u] = d

    return new_ids


def mask_by_domain(s, z, r, t, ri, domain_ids):
    """Splits inputs by predicted domain id

    Args:
        s (_type_): _description_
        z (_type_): _description_
        r (_type_): _description_
        t (_type_): _description_
        ri (_type_): _description_
        d_mask (_type_): _description_
    """

    domain_inputs = []
    domain_dict = {}
    unique_ids = domain_ids[domain_ids.nonzero()].unique()
    for u in unique_ids:
        domain_mask = domain_ids == u
        domain_inputs.append(
            (
                s[:, domain_mask],
                z[:, domain_mask][:, :, domain_mask],
                r[:, domain_mask],
                t[:, domain_mask],
                ri[:, domain_mask],
            )
        )

        domain_dict[ri[0, 0]] = 1

    return domain_inputs, domain_dict


def instance_matrix(labels):
    """Calculate the domain map given some labels.
    Note: using non-prime numbers for the label ids will lead to some
    false domains - we first need to map the labels to primes and then
    map back
    """

    labels_prime = PRIMES.to(labels.device)[labels.long()]

    R = torch.sqrt(torch.einsum("i,j->ij", labels, labels))
    R_prime = torch.sqrt(torch.einsum("i,j->ij", labels_prime, labels_prime))

    R_prime[R_prime % 1 != 0] = 0

    non_domain = R_prime.clone()
    non_domain[non_domain != PRIMES[0]] = 0
    R_prime[R_prime == PRIMES[0]] = 0

    R_clamped = torch.clamp(R_prime.clone(), min=0, max=1)
    R = R * R_clamped

    R_clamped[non_domain == PRIMES[0]] = 0  # 0.5

    return R_clamped, R  # .fill_diagonal_(1), R


def clean_domains(dom_ids: torch.tensor, min_num: int) -> torch.tensor:
    """ Divides any domains that are too small, equally into the domain before and after it.
    """

    dom_counts = torch.unique(dom_ids[dom_ids.nonzero()], return_counts=True)
    remove_ids = dom_counts[0][dom_counts[1] < min_num]

    if len(remove_ids.size()) != 0:
        for i in range(len(dom_ids)):
            if dom_ids[i].item() in remove_ids and i > 0:
                if i != 0:
                    dom_ids[i] = dom_ids[i - 1]

    return dom_ids


def assimilate_short_terminal(dom_ids, dom_counts, threshold, termini):
    """ Assimilates any short stretches of residues the N or C terminus into the 
    proceeding/preceeding domain.
    """
    if termini == 'C':
        dom_counts = dom_counts.flip([0])
        dom_ids = dom_ids.flip([0])

    for i, (c, d) in enumerate(zip(dom_counts, dom_ids)):
        if c >= threshold or d == 0:
            next_dom_id = dom_ids[i + 1]
            break

    dom_ids[:i + 1] = next_dom_id

    if termini == 'C':
        return dom_ids.flip([0])
    else:
        return dom_ids


def get_segment_length(dom_ids):
    """ Returns the length of each segment of a domain.
        e.g. [1,1,1,1,1,1,5,5,5,5,5,2,2,2,2,2] ->
             [6,6,6,6,6,6,5,5,5,5,5,5,5,5,5,5]
    """
    dom_counts = torch.ones_like(dom_ids)
    for i, d in enumerate(dom_ids):
        if i == 0:
            counter = 1
            _idx = idx = i
        else:
            if d == dom_ids[i - 1]:
                counter += 1
                idx = i
            else:
                dom_counts[_idx: idx + 1] = counter
                counter = 0
                _idx = i

        if i == len(dom_ids) - 1:
            dom_counts[_idx: idx + 1] = counter

    return dom_counts


def clean_singletons(dom_ids: torch.tensor, threshold: int) -> torch.tensor:
    """ Re-assigns short segments:
        N-terminus: assimilated into first domain
        C-terminus: assimilated into last domain
        internal segments: divided between the preceeding and proceeding domain
    """

    dom_ids_ = dom_ids
    dom_counts = get_segment_length(dom_ids)

    if len(dom_counts <= threshold) != 0:
        # Assimilate short N-terminal stretches
        dom_ids_ = assimilate_short_terminal(
            dom_ids_, dom_counts, threshold, termini='N')

        # Assimilate short C-terminal stretches
        dom_ids_ = assimilate_short_terminal(
            dom_ids_, dom_counts, threshold, termini='C')

        # Divide internal fragments that are too short
        short_ndr = (dom_counts < threshold)  # * (dom_ids == 0)
        non_terminal = torch.where(short_ndr.long() == 0)[0]

        # Trim off any N and C-terminal ndr stretches
        short_ndr[:non_terminal[0]] = False
        short_ndr[non_terminal[-1]:] = False

        # Subdivide internal NDRs between domains
        true_indices = torch.nonzero(short_ndr).flatten().tolist()

        stretches = []
        for _, g in groupby(enumerate(true_indices), lambda ix: ix[0] - ix[1]):
            stretch = [x[1] for x in g]
            stretches.append(torch.tensor(stretch))

        # Print the stretches
        for stretch in stretches:
            start, end = stretch[0].item(), stretch[-1].item()
            mid = torch.median(stretch).item()

            dom_ids_[start:mid + 1] = dom_ids_[start - 1]
            dom_ids_[mid:end + 1] = dom_ids_[end + 1]

    return dom_ids_


def separate_components(feature_dict: dict, distance: float = 8.0) -> torch.tensor:
    """ Separates domains in the domain map based on a minimum distance.
        Takes the intersect between the domain map and distance map 
        graphs to disconnect discontinuous associations between segments
        when the distance (Angstroms) between them is greater then 
        <distance>. 

    Args:
        domain_map (torch.tensor):  [N,N]   domain map [0,1]
        z (torch.tensor):           [N,N,1] distance map [0,inf]
        domain_ids (torch.tensor):  [N]     domain indices [0, ndom]
        distance (float, optional):         distance cutoff in Angstroms    

    Returns:
        torch.tensor:               [N]     updated domain indices [0, ndom]
    """

    domain_map, z, domain_ids = feature_dict['domain_map'], feature_dict['z'], feature_dict['domain_ids']

    # Re-assign indices based on distance cut off between domains
    dm = z.reshape(z.shape[1], z.shape[1])

    # Convert domain map and thresholded distance map into graphs
    # and compute intersect between both, to remove any vertices
    # in the domain map where the distance is greater than some threshold

    G_int = nx.intersection(
        nx.from_numpy_array(domain_map.detach().cpu().numpy()),
        nx.from_numpy_array((dm <= distance).bool().detach().cpu().numpy()),
    )

    G_int.remove_nodes_from(list(nx.isolates(G_int)))

    # Iterate over each component and re-assign a domain index
    components = list(nx.connected_components(G_int))
    labels_new = torch.zeros_like(domain_ids)

    for i, c in enumerate(components):
        labels_new[list(c)] = i + 1

    return labels_new.to(z.device)


def NonLinCdict(steps, hexcol_array):
    cdict = {"red": (), "green": (), "blue": ()}
    for s, hexcol in zip(steps, hexcol_array):
        rgb = mc.hex2color(hexcol)
        cdict["red"] = cdict["red"] + ((s, rgb[0], rgb[0]),)
        cdict["green"] = cdict["green"] + ((s, rgb[1], rgb[1]),)
        cdict["blue"] = cdict["blue"] + ((s, rgb[2], rgb[2]),)

    return cdict


def format_dom_str(domain_ids: torch.tensor, ri: torch.tensor) -> list:
    """Formats predicted domain ids into residue ranges.
        Domains delimited by ,
        Discontinuous segments delimited by _

        e.g. [0,1,2,3,4,10,11,12] -> ['0-4_10-12']

    Args:
        dom_ids (torch.tensor): predicted domain ids.
        ri (torch.tensor):      residue indices, for mapping back to
                                pdb residues.

    Returns:
        list: _description_     list of domain strings.
    """

    dom_str = []
    for d in torch.unique(domain_ids[domain_ids != 0], sorted=False):
        dom_resi = ri[0, domain_ids == d].long().cpu().tolist()

        strs = []
        for k, g in groupby(enumerate(dom_resi), lambda ix: ix[0] - ix[1]):
            consecs = list(map(itemgetter(1), g))
            if len(consecs) > 1:
                strs.append("{}-{}".format(str(consecs[0]), str(consecs[-1])))
            else:
                strs.append(str(consecs[0]))

        dom_str.append("_".join(strs))

    return ",".join(natsorted(dom_str))



import numpy as np


def mse_loss_at_residue(V_hat, Y, residue_index):
    # loss contribution of residue given current assignment
    # assumes Y is symmetric
    d = residue_index
    # avoid double-counting the diagonal
    return 2*((V_hat[d]@V_hat.T - Y[d])**2).sum() - (V_hat[d]@V_hat[d] - Y[d,d])**2


def null_mse_loss_at_residue(Y, residue_index):
    # loss contribution of residue when unassigned
    d = residue_index
    return 2*(Y[d]**2).sum() - Y[d,d]**2


def mse_loss(V_hat, Y):
    # only use to compute initial loss
    return ((V_hat@V_hat.T - Y)**2).sum()


def nll_loss(V_hat, Y):
    """We now interpret ys as probabilities.

    We want to minimise the negative log probability.
    
    NB Y must be symmetric and [0-1]
    """
    # only use to compute initial loss
    log_pij = (V_hat@V_hat.T) * np.log(Y) + (1-(V_hat@V_hat.T))*np.log(1-Y)
    # simple sum double counts off diagonals but single counts on diagonals, so need
    # to add diagonal again before dividing by 2
    return -0.5*(log_pij.sum() + np.diag(log_pij).sum())


def nll_loss_at_residue(V_hat, Y, residue_index):
    # d is residue_index
    d = residue_index  # V_hat[d]@V_hat.T is just Adj without instantiating A
    # n.b. whereas in case of mse we double-count off-diagonals, here we don't, hence the difference
    log_pdj = (V_hat[d]@V_hat.T) * np.log(Y[d]) + (1-(V_hat[d]@V_hat.T))*np.log(1-Y[d])
    return -log_pdj.sum()


def null_nll_loss_at_residue(Y, residue_index):
    return - np.log(1-Y[residue_index]).sum()


def greedy_V(Y, N_iters=3, K_init=4, cost_type="mse"):
    """ 
    Learn a binary matrix V, with at most one nonzero entry per row, to minimize
    
     || VV' - Y ||_2
     
    This is done by initializing a V of all zeros, and then doing a greedy optimization.
    V is initially D x K_init, where D is the number of residues (i.e., where Y is DxD).
    K is learned automatically; different values of K_init will not change the result but
    might be more or less efficient in terms of memory usage. The code keeps track of the 
    current number of nonzero columns of V, and when all the columns are full it adds extra
    columns of zeros at the end.
    
    Each iteration sweeps through all residues (i.e. rows of V) once.

    The implementation relies on additivity of loss to reduce computational cost
    by only computing required increments to the loss at each iteration.

    
    INPUTS:
    
    Y: matrix of predictions with entries in [0, 1]
    N_iters: number of iterations
    K_init: initial number of columns of V, adjust this for tweaking performance
    """
    Y = (Y + Y.T) / 2 # required for consistency with model_v1
    V_hat = np.zeros((Y.shape[0], K_init), dtype=np.uint8)
    if cost_type == "mse":
        loss = mse_loss(V_hat, Y) # initial loss for the zero matrix V_hat = 0
    elif cost_type == "nll":
        loss = nll_loss(V_hat, Y)
    else:
        raise ValueError(cost_type)
    K_max = K_init # track number of columns in K_max

    for it in range(N_iters):
        for d in range(V_hat.shape[0]):
            # d is a residue index
            # compute the loss, excluding contribution of this residue
            # TODO check minus signs are consistent
            loss_minus_d = loss - mse_loss_at_residue(V_hat, Y, d)

            # sweep through all K+1 options and compute what the contribution to the loss would be
            V_hat[d] *= 0

            # loss with no assignment
            L0 = loss_minus_d + null_mse_loss_at_residue(Y, d)
            L_opt = np.zeros(K_max)
            for k in range(K_max):
                # note this could be vectorized with a bit of work
                V_hat[d,k] = 1
                L_opt[k] = loss_minus_d + mse_loss_at_residue(V_hat, Y, d)
                V_hat[d,k] = 0

            # select the option which minimizes the squared error
            z = np.argmin(L_opt)
            if L_opt[z] < L0:
                V_hat[d,z] = 1
            
            # update loss
            loss = loss_minus_d + mse_loss_at_residue(V_hat, Y, d)

            if z == K_max-1:
                # Expand V_hat to make room for extra potential clusters (no nonzero columns remain)
                # TODO: note that if a cluster is "removed" on a later iteration we will miss this.
                #       probably doesn't matter, just means it is occasionally possible that the V
                #       that is returned could sometimes have a few nonzero columns.
                V_hat = np.concatenate((V_hat, np.zeros_like(V_hat)), -1)
                K_max = V_hat.shape[1]

    # drop columns that are zeros (i.e. unused clusters) before returning
    empty = V_hat.sum(0) == 0
    return V_hat[:,~empty], loss



import unittest

from src import MerizoCluster


class TestProcessInput(unittest.TestCase):
    def test_merizo(self):
        expected_output = "1-71,72-143"

        actual_output = MerizoCluster().predict_from_pdb('prot.pdb')

        self.assertEqual(actual_output, expected_output, "Merizo output does not match expected")



"""
Domain boundary distance score, as defined in CASP 7 paper:
Assessment of predictions submitted for the CASP7 domain prediction category (2007)
https://onlinelibrary.wiley.com/doi/10.1002/prot.21675

Under our scoring scheme all predictions within 8 residues of the correct boundary
will score, but predictions that are closer to the correct domain boundary would score more.
All distances between the predicted and correct domain boundaries are calculated. If the
domain boundary has a linker, the whole linker is regarded as the domain boundary.
predictions are given one point for being within 1 residue of each correct boundary,
another point if they are within two residues, a further point if they are within three,
and so on up to eight residues. A prediction two residues away from the correct boundary
would therefore have 7 points.
The total score for each domain prediction is then calculated as the sum of all predicted
boundary scores divided by eight and the total number of domain boundaries. The number of
domain boundaries comes from either the target or the prediction, whichever is higher. In
this way over-prediction is penalized.

"""
import numpy as np


def pred_domains_to_bounds(pred_domains, optimize_for_linkers=False):
    """
    Converts domain dictionary to list of boundary residues
    """
    pred_bounds = np.array([])
    for name, res in pred_domains.items():
        if name=='linker' or len(res) ==0:
            continue
        res = np.array(sorted(set(res)))
        gaps  = res[1:] - res[:-1]
        if any(gaps > 1):
            gaps_start_indexes = np.where(gaps > 1)[0]
            pred_bounds = np.append(pred_bounds, res[gaps_start_indexes] + 1)
            pred_bounds = np.append(pred_bounds, res[gaps_start_indexes + 1])
        pred_bounds = np.append(pred_bounds, [res[0], res[-1] +1])
    if optimize_for_linkers:
        # given that linkers all count as boundaries probs possible
        # to score more points by predicting boundaries in the middle of linker
        # regions rather than the edges
        raise NotImplementedError # todo
    return np.array(sorted(list(set(pred_bounds.astype(int)))))


def get_true_boundary_res(domain_dict):
    """
    In the case where there are multiple non-domain
    residues between two domains, all of these NDRs
    are counted as domain boundaries. This adjustment
    is applied to the true boundaries but not the
    predicted boundaries.
    """
    bounds = pred_domains_to_bounds(domain_dict)
    # c.f. get_boundary_res below
    boundaries = {
        "boundary_res": list(bounds),
        "n_boundaries": len(list(bounds))
    }
    boundaries["boundary_res"] += list(domain_dict["linker"])
    return boundaries


def boundary_distance_score(domains, boundaries):
    pred_bounds = pred_domains_to_bounds(domains)
    # distance score as specified in CASP7 paper {distance_to_true_bound:score}
    dist_to_score = {0:8, 1:7, 2:6, 3:5, 4:4, 5:3, 6:2, 7:1} 
    score = 0
    # the final score is divided by the number of segments and the maximum score for each boundary
    normalizing_term = 8 * max(len(pred_bounds), boundaries['n_boundaries'])
    scores = []
    for b in pred_bounds:
        distance = min(abs(boundaries['boundary_res'] - b))
        if distance < 8:
            scores.append(dist_to_score[distance])
    # JW adjustment: additional boundaries that can't be mapped to a real bound 
    # should not be added to the un-normalized score
    score = sum(sorted(scores)[-boundaries['n_boundaries']:])
    return score / normalizing_term



class BaseError(Exception):
    """
    Error class for all local exceptions in this code base
    """
    pass

class PredictionResultExistsError(BaseError):
    pass

class FileExistsError(BaseError):
    pass


from torch import nn, transpose


def elu():
    return nn.ELU(inplace=True)


def instance_norm(filters, eps=1e-6, **kwargs):
    return nn.InstanceNorm2d(filters, affine=True, eps=eps, **kwargs)


def conv2d(in_chan, out_chan, kernel_size, dilation=1, **kwargs):
    padding = dilation * (kernel_size - 1) // 2
    return nn.Conv2d(in_chan, out_chan, kernel_size, padding=padding, dilation=dilation, **kwargs)


class trRosettaNetwork(nn.Module):
    def __init__(self, filters=64, kernel=3, num_layers=61, in_channels=3, symmetrise_output=False, dropout=0.15):
        super().__init__()
        self.filters = filters
        self.kernel = kernel
        self.num_layers = num_layers
        self.in_channels = in_channels
        self.symmetrise_output = symmetrise_output
        self.first_block = nn.Sequential(
            conv2d(self.in_channels, filters, 1),
            instance_norm(filters),
            elu()
        )
        self.output_layer = nn.Sequential(
            conv2d(filters, 1, kernel, dilation=1),
            nn.Sigmoid())


        # stack of residual blocks with dilations
        cycle_dilations = [1, 2, 4, 8, 16]
        dilations = [cycle_dilations[i % len(cycle_dilations)] for i in range(num_layers)]
        if dropout > 0:
            self.layers = nn.ModuleList([nn.Sequential(
                conv2d(filters, filters, kernel, dilation=dilation),
                instance_norm(filters),
                elu(),
                nn.Dropout(p=dropout),
                conv2d(filters, filters, kernel, dilation=dilation),
                instance_norm(filters)
            ) for dilation in dilations])
        else:
            self.layers = nn.ModuleList([nn.Sequential(
                conv2d(filters, filters, kernel, dilation=dilation),
                instance_norm(filters),
                elu(),
                conv2d(filters, filters, kernel, dilation=dilation),
                instance_norm(filters)
            ) for dilation in dilations])

        self.activate = elu()


    def forward(self, x):
        x = self.first_block(x)

        for layer in self.layers:
            x = self.activate(x + layer(x))
        y_hat = self.output_layer(x)
        if self.symmetrise_output:
            return (y_hat + transpose(y_hat, -1, -2)) * 0.5
        else:
            return y_hat



import torch
import torch.nn.functional as F
from torch import nn

from src.merizo.model.utils.utils import clean_domains, clean_singletons

# https://github.com/rstrudel/segmenter/blob/master/segm/model/decoder.py

class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim, dropout, out_dim=None):
        super().__init__()
        self.fc1 = nn.Linear(dim, hidden_dim)
        self.act = nn.GELU()
        if out_dim is None:
            out_dim = dim
        self.fc2 = nn.Linear(hidden_dim, out_dim)
        self.drop = nn.Dropout(dropout)

    @property
    def unwrapped(self):
        return self

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, heads, dropout):
        super().__init__()
        self.heads = heads
        head_dim = dim // heads
        self.attn = None

        self.qkv = nn.Linear(dim, dim * 3)
        self.attn_drop = nn.Dropout(dropout)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(dropout)

    @property
    def unwrapped(self):
        return self

    def forward(self, x, mask=None, bias=None):
        B, N, C = x.shape
        qkv = (
            self.qkv(x)
            .reshape(B, N, 3, self.heads, C // self.heads)
            .permute(2, 0, 3, 1, 4)
        )

        q, k, v = (qkv[0], qkv[1], qkv[2])
        qk = q @ k.transpose(-2, -1)

        if bias is not None:
            qk = qk + bias

        attn = qk 
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x, attn


class Block(nn.Module):
    def __init__(self, dim, heads, mlp_dim, dropout, drop_path):
        super().__init__()
        self.norm1 = nn.LayerNorm(dim)
        self.norm2 = nn.LayerNorm(dim)
        self.attn = Attention(dim, heads, dropout)
        self.mlp = FeedForward(dim, mlp_dim, dropout)
        self.drop_path = nn.Identity()

    def forward(self, x, mask=None, return_attention=False, bias=None):
        y, attn = self.attn(self.norm1(x), mask, bias)
        if return_attention:
            return attn
        x = x + self.drop_path(y)
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class MaskTransformer(nn.Module):
    def __init__(
        self,
        n_cls,
        n_layers,
        n_heads,
        d_model,
        d_ff,
        drop_path_rate,
        dropout,
    ):
        super().__init__()
        self.n_layers = n_layers
        self.n_cls = n_cls
        self.d_model = d_model
        self.d_ff = d_ff
        self.scale = d_model**-0.5

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, n_layers)]
        self.blocks = nn.ModuleList(
            [Block(d_model, n_heads, d_ff, dropout, dpr[i]) for i in range(n_layers)]
        )

        self.cls_emb = nn.Parameter(torch.randn(1, n_cls, d_model))

        self.proj_patch = nn.Parameter(self.scale * torch.randn(d_model, d_model))
        self.proj_classes = nn.Parameter(self.scale * torch.randn(d_model, d_model))

        self.decoder_norm = nn.LayerNorm(d_model)
        self.class_norm = nn.LayerNorm(n_cls)

        # Background prediction layers
        self.bg_gru = nn.GRU(
            input_size=d_model,
            hidden_size=d_model // 2,
            batch_first=True,
            num_layers=2,
            dropout=0,
            bidirectional=True,
        )

        self.bg_out = nn.Linear(d_model, 2)

        # Confidence Layers
        self.conf_gru = nn.GRU(
            input_size=n_cls,
            hidden_size=d_model,
            batch_first=True,
            num_layers=2,
            dropout=0,
            bidirectional=True,
        )

        self.conf_gru_all = nn.GRU(
            input_size=n_cls,
            hidden_size=d_model,
            batch_first=True,
            num_layers=2,
            dropout=0,
            bidirectional=True,
        )

        self.conf_out_all = nn.Linear(d_model, 1)
        self.conf_out = nn.Linear(d_model, 1)

    def forward(self, s, bias):

        # Initialise tensor for cls embeddings
        cls_emb = self.cls_emb.expand(s.size(0), -1, -1)

        # Concat to input at dim 1
        x = torch.cat((s, cls_emb), 1)  # , bg_emb), 1)

        # Alibi
        if bias is not None:
            pad = self.n_cls  # + 2
            bias = F.pad(bias, (0, pad, 0, pad), "constant", 0)

        # MHA blocks
        for blk in self.blocks:
            x = blk(x, bias)

        x = self.decoder_norm(x)

        # separate out original input patches and cls embeddings
        features = x[:, : -self.n_cls] @ self.proj_patch
        classes = x[:, -self.n_cls :] @ self.proj_classes

        # Some normalisation
        features = features / features.norm(dim=-1, keepdim=True)
        classes = classes / classes.norm(dim=-1, keepdim=True)

        # Dot product of (1,nres,emb) . (1,nclass,emb) = (1,nres,nclass)
        domain_masks = self.class_norm(features @ classes.transpose(1, 2))
        bg_masks = self.bg_out(self.bg_gru(features)[0])

        # Get per-domain confidence based on predictions
        pred_ids = domain_masks.argmax(dim=-1).flatten()

        # Cleaning before x bg
        pred_ids = clean_domains(pred_ids, 50)
        pred_ids = clean_singletons(pred_ids, 10)

        # Apply bg residue mask
        dom_ids = pred_ids * bg_masks.argmax(dim=-1).flatten()

        # Get unique domain indices
        unique_ids = dom_ids[dom_ids.nonzero()].unique()

        # Iterate over each pred index and pass through gru to predict confidence
        conf_res = torch.zeros_like(dom_ids).float()
        
        for i, d in enumerate(unique_ids):
            dom = domain_masks[:, dom_ids == d]

            dom_conf = (
                self.conf_out(self.conf_gru(dom)[1][-1:, :, :])
                .flatten()
                .clamp(min=0, max=1)
            )

            conf_res[dom_ids == d] = dom_conf

        return dom_ids.reshape(-1), conf_res.reshape(-1)
    



"""Domain predictor classes.
"""
import os
import glob
import numpy as np
import pandas as pd
import torch
from torch import nn
from src.chainsaw.domain_assignment.util import make_pair_labels, make_domain_mapping_dict

import logging
LOG = logging.getLogger(__name__)


def get_checkpoint_epoch(checkpoint_file):
    return int(os.path.splitext(checkpoint_file)[0].split(".")[-1])


class PairwiseDomainPredictor(nn.Module):

    """Wrapper for a pairwise domain co-membership predictor, adding in domain prediction post-processing."""

    def __init__(
        self,
        model,
        domain_caller,
        device,
        loss="bce",
        x_has_padding_mask=True,
        mask_padding=True,
        n_checkpoints_to_average=1,
        checkpoint_dir=None,
        load_checkpoint_if_exists=False,
        save_val_best=True,
        max_recycles=0,
        post_process_domains=True,
        min_ss_components=2,
        min_domain_length=30,
        remove_disordered_domain_threshold=0,
        trim_each_domain=True,
        dist_transform_type="min_replace_inverse",
        distance_denominator=10,
    ):
        super().__init__()
        self._train_model = model  # we want to keep this hanging around so that optimizer references dont break
        self.model = self._train_model
        self.domain_caller = domain_caller
        self.device = device
        self.x_has_padding_mask = x_has_padding_mask
        self.mask_padding = mask_padding  # if True use padding mask to mask loss
        self.n_checkpoints_to_average = n_checkpoints_to_average
        self.checkpoint_dir = checkpoint_dir
        self._epoch = 0
        self.save_val_best = save_val_best
        self.best_val_metrics = {}
        self.max_recycles = max_recycles
        self.post_process_domains = post_process_domains
        self.remove_disordered_domain_threshold = remove_disordered_domain_threshold
        self.trim_each_domain = trim_each_domain
        self.min_domain_length = min_domain_length
        self.min_ss_components = min_ss_components
        self.dist_transform_type = dist_transform_type
        self.distance_denominator = distance_denominator
        if load_checkpoint_if_exists:
            checkpoint_files = sorted(
                glob.glob(os.path.join(self.checkpoint_dir, "weights*")),
                key=get_checkpoint_epoch,
                reverse=True,
            )
            if len(checkpoint_files) > 0:
                self._epoch = get_checkpoint_epoch(checkpoint_files[0])
                LOG.info(f"Loading saved checkpoint(s) ending at epoch {self._epoch}")
                self.load_checkpoints(average=True)
                self.load_checkpoints()
            else:
                LOG.info("No checkpoints found to load")

        if loss == "bce":
            self.loss_function = nn.BCELoss(reduction="none")
        elif loss == "mse":
            self.loss_function = nn.MSELoss(reduction="none")

    def load_checkpoints(self, average=False, old_style=False):
        if self.n_checkpoints_to_average == 1:
            weights_file = os.path.join(self.checkpoint_dir, "weights.pt")
        else:
            # for e.g. resetting training weights for next training epoch after testing with avg
            LOG.info(f"Loading last checkpoint (epoch {self._epoch})")
            weights_file = os.path.join(self.checkpoint_dir, f"weights.{self._epoch}.pt")
        LOG.info(f"Loading weights from: {weights_file}")
        state_dict = torch.load(weights_file, map_location=self.device)
        if old_style:
            self.load_state_dict(state_dict, strict=False)
        else:
            self.model.load_state_dict(state_dict)

    def predict_pairwise(self, x):
        x = x.to(self.device)
        if np.isnan(x.cpu().numpy()).any():
            raise Exception('NAN values in data')
        y_pred = self.model(x).squeeze(1)  # b, L, L
        assert y_pred.ndim == 3
        return y_pred

    def get_mask(self, x):
        """Binary mask 1 for observed, 0 for padding."""
        x = x.to(self.device)
        if self.x_has_padding_mask:
            mask = 1 - x[:, -1]  # b, L, L
        else:
            mask = None
        return mask

    def epoch_start(self):
        self.model = self._train_model
        self.model.train()
        self._epoch += 1

    def test_begin(self):
        if self.n_checkpoints_to_average > 1:
            torch.save(self.model.state_dict(), os.path.join(self.checkpoint_dir, f'weights.{self._epoch}.pt'))
            start_idx = self._epoch - self.n_checkpoints_to_average
            if start_idx >= 2:
                os.remove(os.path.join(self.checkpoint_dir, f"weights.{start_idx-1}.pt"))
        else:
            torch.save(self.model.state_dict(), os.path.join(self.checkpoint_dir, 'weights.pt'))

        if self.n_checkpoints_to_average > 1:
            # self.model.to("cpu")  # free up gpu memory for average model
            self.load_checkpoints(average=True)

        self.model.eval()

    def forward(self, x, y, batch_average=True):
        """A training step."""
        x, y = x.to(self.device), y.to(self.device)
        y_pred = self.predict_pairwise(x)
        mask = self.get_mask(x)
        return self.compute_loss(y_pred, y, mask=mask)

    def compute_loss(self, y_pred, y, mask=None, batch_average=True):
        y_pred, y = y_pred.to(self.device), y.to(self.device)
        if mask is None or not self.mask_padding:
            mask = torch.ones_like(y)
        # mask is b, L, L. To normalise correctly, we need to divide by number of observations
        loss = (self.loss_function(y_pred, y)*mask).sum((-1,-2)) / mask.sum((-1,-2))

        # metrics characterising inputs: how many residues, how many with domain assignments.
        labelled_residues = ((y*mask).sum(-1) > 0).sum(-1)  # b
        non_padding_residues = (mask.sum(-1) > 0).sum(-1)  # b
        labelled_frac = labelled_residues / non_padding_residues
        metrics = {
            "labelled_residues": labelled_residues.detach().cpu().numpy(),
            "residues": non_padding_residues.detach().cpu().numpy(),
            "labelled_frac": labelled_frac.detach().cpu().numpy(),
            "loss": loss.detach().cpu().numpy(),
        }
        if batch_average:
            loss = loss.mean(0)
            metrics = {k: np.mean(v) for k, v in metrics.items()}

        return loss, metrics

    def domains_from_pairwise(self, y_pred):
        assert y_pred.ndim == 3
        domain_preds = []
        confidence_list = []
        for pred_single in y_pred.cpu().numpy():
            single_domains, confidence = self.domain_caller(pred_single)
            domain_preds.append(single_domains)
            confidence_list.append(confidence)
        return domain_preds, confidence_list

    def distance_transform(self, x):
        dist_chan = x[0, 0]
        # Find the minimum non-zero value in the channel
        min_nonzero = dist_chan[dist_chan > 0].min()
        # Replace zero values in the channel with the minimum non-zero value
        dist_chan[dist_chan == 0] = min_nonzero
        if self.dist_transform_type == "min_replace_inverse":
            # replace zero values and then invert.
            dist_chan = dist_chan ** (-1)
            x[0, 0] = dist_chan
            return x

        elif self.dist_transform_type == "unidoc_exponent": # replace zero values in pae / distance
            spread = self.distance_denominator
            dist_chan = (1 + np.exp((dist_chan - 8) / spread)) ** -1
            x[0,0] = dist_chan
            return x

    @torch.no_grad()
    def predict(self, x, return_pairwise=True):
        x = self.distance_transform(x)
        if self.max_recycles > 0:
            for i in range(self.max_recycles):
                # add recycling channels
                n_res = x.shape[-1]
                recycle_channels = torch.zeros(1, 2, n_res, n_res)
                # Concatenate the original tensor and the zeros tensor along the second dimension
                x = torch.cat((x, recycle_channels), dim=1)
                x = self.recycle_predict(x)
        y_pred = self.predict_pairwise(x)
        domain_dicts, confidence = self.domains_from_pairwise(y_pred)
        if self.post_process_domains:
            domain_dicts = self.post_process(domain_dicts, x) # todo move this to domains from pairwise function
        if return_pairwise:
            return y_pred, domain_dicts, confidence
        else:
            return domain_dicts, confidence

    @torch.no_grad()
    def recycle_predict(self, x):
        x = x.to(self.device)
        y_pred = self.predict_pairwise(x)
        domain_dicts, confidence = self.domains_from_pairwise(y_pred)
        y_pred_from_domains = np.array(
            [make_pair_labels(n_res=x.shape[-1], domain_dict=d_dict) for d_dict in domain_dicts])
        y_pred_from_domains = torch.tensor(y_pred_from_domains).to(self.device)
        if self.x_has_padding_mask:
            x[:, -2, :, :] = y_pred # assumes that last dimension is padding mask
            x[:, -3, :, :] = y_pred_from_domains
        else:
            x[:, -1, :, :] = y_pred
            x[:, -2, :, :] = y_pred_from_domains
        return x


    def post_process(self, domain_dicts, x_batch):
        new_domain_dicts = []
        for domain_dict, x in zip(domain_dicts, x_batch):
            x = x.cpu().numpy()
            domain_dict = {k: list(v) for k, v in domain_dict.items()}
            helix, sheet = x[1], x[2]
            diag_helix = np.diagonal(helix)
            diag_sheet = np.diagonal(sheet)
            ss_residues = list(np.where(diag_helix > 0)[0]) + list(np.where(diag_sheet > 0)[0])

            domain_dict = self.trim_disordered_boundaries(domain_dict, ss_residues)

            if self.remove_disordered_domain_threshold > 0:
                domain_dict = self.remove_disordered_domains(domain_dict, ss_residues)

            if self.min_ss_components > 0:
                domain_dict = self.remove_domains_with_few_ss_components(domain_dict, x)

            if self.min_domain_length > 0:
                domain_dict = self.remove_domains_with_short_length(domain_dict)
            new_domain_dicts.append(domain_dict)
        return new_domain_dicts

    def trim_disordered_boundaries(self, domain_dict, ss_residues):
        if not self.trim_each_domain:
            start = min(ss_residues)
            end = max(ss_residues)
        for dname, res in domain_dict.items():
            if dname == "linker":
                continue
            if self.trim_each_domain:
                domain_specific_ss = set(ss_residues).intersection(set(res))
                if len(domain_specific_ss) == 0:
                    continue
                start = min(domain_specific_ss)
                end = max(domain_specific_ss)
            domain_dict["linker"] += [r for r in res if r < start or r > end]
            domain_dict[dname] = [r for r in res if r >= start and r <= end]
        return domain_dict

    def remove_disordered_domains(self, domain_dict, ss_residues):
        new_domain_dict = {}
        for dname, res in domain_dict.items():
            if dname == "linker":
                continue
            if len(res) == 0:
                continue
            if len(set(res).intersection(set(ss_residues))) / len(res) < self.remove_disordered_domain_threshold:
                domain_dict["linker"] += res
            else:
                new_domain_dict[dname] = res
        new_domain_dict["linker"] = domain_dict["linker"]
        return new_domain_dict


    def remove_domains_with_few_ss_components(self, domain_dict, x):
        """
        Remove domains where number of ss components is less than minimum
        eg if self.min_ss_components=2 domains made of only a single helix or sheet are removed
        achieve this by counting the number of unique string hashes in domain rows of x
        """
        new_domain_dict = {}
        for dname, res in domain_dict.items():
            if dname == "linker":
                continue
            res = sorted(res)
            helix = x[1][res, :][:, res]
            strand = x[2][res, :][:, res]
            helix = helix[np.any(helix, axis=1)]
            strand = strand[np.any(strand, axis=1)]
            # residues in the same secondary structure component have the same representation in the helix or strand matrix
            n_helix = len(set(["".join([str(int(i)) for i in row]) for row in helix]))
            n_sheet = len(set(["".join([str(int(i)) for i in row]) for row in strand]))
            if len(res) == 0:
                continue
            if n_helix + n_sheet < self.min_ss_components:
                domain_dict["linker"] += res
            else:
                new_domain_dict[dname] = res
        new_domain_dict["linker"] = domain_dict["linker"]
        return new_domain_dict

    def remove_domains_with_short_length(self, domain_dict):
        """
        Remove domains where length is less than minimum
        """
        new_domain_dict = {}
        for dname, res in domain_dict.items():
            if dname == "linker":
                continue


            if len(res) < self.min_domain_length:
                domain_dict["linker"] += res
            else:
                new_domain_dict[dname] = res
        new_domain_dict["linker"] = domain_dict["linker"]
        return new_domain_dict


class CSVDomainPredictor:
    def __init__(self, csv_predictions):
        self.csv_filepath = csv_predictions
        self.predictions = pd.read_csv(csv_predictions)

    def predict(self, x):
        # x should be a chain_id
        one_pred = self.predictions[self.predictions.chain_id == x[0][:5]]
        if len(one_pred) == 0:
            return None
        domain_dicts = [make_domain_mapping_dict(one_pred.iloc[0])]
        return None, domain_dicts



import torch
import torch.nn
import numpy as np

from .pdb_parser import get_xyz, select_from_mol


def get_frames(pdb, missing=None):
    """ Function for calculating "frames" for N-CA-C atoms of a list of residues
        Calibrated against 'rigids_from_3_points' in AF2 r3.py (Jumper et al., 2021, Nature)
    Args:
        pdb: pdb in a numpy structured array format
    Returns:
        (R,t): R - (nres,3,3) rotation matrix
               t - (nres,3) translation matrix
    """

    if missing and len(missing) != 0:
        missing_pad = []
        for m in missing:
            for n in ['N', 'CA', 'C']:
                missing_pad.append(
                    ('ATOM', 0, n, '', 'PAD', 'A',
                     m, 0., 0., 0., 0., pdb[0]['b'][0])
                )

        missing_pad = np.array(
            missing_pad,
            dtype=[('type', 'U6'), ('i', 'i4'), ('n', 'U4'), ('alt', 'U1'),
                   ('resn', 'U3'), ('chain', 'U2'), ('resi', 'i4'), ('x', 'f8'),
                   ('y', 'f8'), ('z', 'f8'), ('occ', 'f8'), ('b', 'f8')]
        )

        pdb = [np.sort(np.concatenate((pdb[0], missing_pad)), order='resi')]

    x1 = get_xyz(select_from_mol(pdb, 'n', ['N'])[0], gaps=False)
    x2 = get_xyz(select_from_mol(pdb, 'n', ['CA'])[0], gaps=False)
    x3 = get_xyz(select_from_mol(pdb, 'n', ['C'])[0], gaps=False)
    R, t = rigidFrom3Points(x1, x2, x3, epsilon=1e-8)

    return (torch.tensor(R), torch.tensor(t))


def rigidFrom3Points(x1, x2, x3, epsilon=1e-8):
    """ Re-written from Jumper et al., 2021 Suppl. Alg. 21 'rigidFrom3Points'
        Calibrated against 'rigids_from_3_points' in r3.py

    [Vectorised version]

    Creates a set of rigid transformations from 3 points using the Gram-Schmidt
    orthonormalisation procedure.
    Args:
        x1: [x,y,z] of N atom
        x2: [x,y,z] of CA atom
        x3: [x,y,z] of C atom
        epsilon: added to normalisation, 1e-8 in paper
    Returns:
        (R,t) where R is the (nres,3,3) rotation matrix and t is the (nres,3) translation matrix
        that compose the rigid transforms needed to transform the original coordinates
        to the basis vectors.

        R consists of [e1,e2,e3] where e1,e2,e3 are 1x3 basis vectors.
        e1 and e2 are in the CA-N and CA-C axis respectively
        e3 is the cross-product of e1 and e2.
    """

    v1 = x2-x1 #x3-x2  # in Supp. Alg. 21: v1=x3-x2 / in r3.py: v1=x2-x1
    v2 = x3-x2 #x1-x2  # in Supp. Alg. 21: v2=x1-x2 / in r3.py: v2=x3-x2
    v1_norm = np.sqrt(
        np.square(v1[0]) + np.square(v1[1]) + np.square(v1[2]) + epsilon)

    e1 = v1/v1_norm  # 1st basis vector

    u2 = v2 - (e1*np.sum(e1*v2, axis=0))
    u2_norm = np.sqrt(
        np.square(u2[0]) + np.square(u2[1]) + np.square(u2[2]) + epsilon)
    e2 = u2/u2_norm  # 2nd basis vector

    e3 = np.cross(e1.T, e2.T)  # 3rd basis vector

    R = np.concatenate(
        (e1.reshape(-1, 3, 1), e2.reshape(-1, 3, 1), e3.reshape(-1, 3, 1)), axis=2)

    return (R, x2.T)



"""
Utilties for converting to/from a dictionary representation of domain assignments.
"""

import logging
import os
from itertools import product

import numpy as np

LOG = logging.getLogger(__name__)


def make_pair_labels(n_res, domain_dict, id_string=None, save_dir=None, non_aligned_residues=[]):
    """n_res: number of residues in the non-trimmed sequence

        non_aligned_residues: these will be used to trim down from n_res

        domain_dict: eg. {'D1': [0,1,2,3], 'D2': [4,5,6]}
    """
    pair_labels = np.zeros([n_res, n_res])
    for domain, res_ix in domain_dict.items():
        if domain == 'linker':
            continue
        coords_tuples = list(product(res_ix, res_ix))
        x_ix = [i[0] for i in coords_tuples]
        y_ix = [i[1] for i in coords_tuples]
        pair_labels[x_ix, y_ix] = 1
    if len(non_aligned_residues):
        aligned_residues = [i for i in range(n_res) if i not in non_aligned_residues]
        pair_labels = pair_labels[aligned_residues,:][:,aligned_residues]
    if save_dir is not None:
        save_path = os.path.join(save_dir, id_string)
        np.savez_compressed(save_path, pair_labels)

    return pair_labels


def sort_domain_limits(limits, dom_names):
    start_positions = [x[0] for x in limits]
    end_positions = [x[1] for x in limits]
    sorted_index = np.argsort(start_positions)
    assert (sorted_index == np.argsort(end_positions)).all()
    return np.array(limits)[sorted_index], list(np.array(dom_names)[sorted_index])


def resolve_residue_in_multiple_domain(mapping, shared_res):
    """
    This is a stupid slow recursive solution: but I think it only applies to one
    case so going to leave it for now
    """
    for one_shared in shared_res:
        for domain, res in mapping.items():
            if one_shared in res:
                mapping[domain].remove(one_shared)
                return check_no_residue_in_multiple_domains(mapping)


def check_no_residue_in_multiple_domains(mapping, resolve_conflics=True):
    # ensures no residue index is associated with more than one domain
    for dom, res in mapping.items():
        for dom2, res2 in mapping.items():
            if dom == dom2:
                continue
            shared_res = set(res).intersection(set(res2))
            if len(shared_res):
                print(f'Found {len(shared_res)} shared residues')
                if resolve_conflics:
                    mapping = resolve_residue_in_multiple_domain(mapping, shared_res)
                else:
                    raise ValueError("SAME RESIDUE NUMBER FOUND IN MULTIPLE DOMAINS")
    return mapping


def make_domain_mapping_dict(row):
    dom_limit_list = row.dom_bounds_pdb_ix.split('|')
    dom_names = row.dom_names.split('|')
    dom_limit_list = convert_limits_to_numbers(dom_limit_list)
    dom_limit_array, dom_names = sort_domain_limits(dom_limit_list, dom_names)
    mapping = {}

    for i, d_lims in enumerate(dom_limit_array):
        dom_name = dom_names[i]
        pdb_start, pdb_end = d_lims
        if dom_name not in mapping:
            mapping[dom_name] = []
        mapping[dom_name] += list(range(pdb_start, pdb_end))
    check_no_residue_in_multiple_domains(mapping)
    return mapping


def convert_limits_to_numbers(dom_limit_list):
    processed_dom_limit_list = []
    for lim in dom_limit_list:
        dash_idx = [i for i, char in enumerate(lim) if char == '-']
        if len(dash_idx) == 1:
            start_index = int(lim.split('-')[0]) -1
            end_index = int(lim.split('-')[1])
        else:
            raise ValueError('Invalid format for domain limits', str(dom_limit_list))
        processed_dom_limit_list.append((start_index, end_index))
    return processed_dom_limit_list


def convert_domain_dict_strings(domain_dict):
    """
    Converts the domain dictionary into domain_name string and domain_bounds string
    eg. domain names D1|D2|D1
    eg. domain bounds 0-100|100-200|200-300
    """
    domain_names = []
    domain_bounds = []
    for k,v in domain_dict.items():
        if k=='linker':
            continue
        residues = sorted(v)
        for i, res in enumerate(residues):
            if i==0:
                start = res
            elif residues[i-1] != res - 1:
                domain_bounds.append(f'{start}-{residues[i-1]}')
                domain_names.append(k)
                start = res
            if i == len(residues)-1:
                domain_bounds.append(f'{start}-{res}')
                domain_names.append(k)

    return '|'.join(domain_names), '|'.join(domain_bounds)



import unittest

from src import ChainsawCluster


class TestChainsaw(unittest.TestCase):

    def test_chainsaw(self):
        expected_output = "3-71,77-143"
        actual_output = ChainsawCluster().predict_from_pdb('prot.pdb')
        self.assertEqual(actual_output, expected_output, "Chainsaw output does not match expected")



from collections import defaultdict
import copy
import numpy as np


def check_unique_assignments(assn):
    res_counts = defaultdict(int)  # check for repeated residues
    for res_ids in assn.values():
        for res_id in res_ids:
            res_counts[res_id] += 1
    assert all([v == 1 for v in res_counts.values()]), f"Non-unique domain assignment {assn}"


def make_domain_dict(domain_str, n_res):
    """
    Converts string representation of domain boundaries to a dictionary
    Delimit separate domains with commas , and discontinuous domains
    with underscores _. Residue ranges separated by hyphens -, e.g. 1-100,101-200_300-340
    :param domain_str: domain boundaries expressed zero-indexed sequential  e.g. 1-100,101-200_300-340
    :param n_res: number of residues in the sequence
    :return:
    """
    domain_dict = {}
    bounds = domain_str.split(',')
    assigned_res = set()
    for i, bound in enumerate(bounds):
        if len(bound):

            for segment in bound.split('_'):
                if '-' in segment:
                    start, end = segment.split('-')
                else:
                    start = end = segment
                segment_res = set(range(int(start), int(end) + 1))
                assert len(segment_res.intersection(assigned_res)) == 0, f"Overlapping domain assignments {domain_str}"
                assigned_res.update(segment_res)
                domain_dict[f"D{i + 1}"] = domain_dict.get(f"D{i + 1}", []) + list(segment_res)
    domain_dict['linker'] = list(set(range(n_res)).difference(assigned_res))
    return domain_dict


def ndo_score(true: dict, pred: dict):
    """
    Normalized Domain Overlap Score
    Approximately corresponds to the fraction of residues that are assigned to the correct domain
    (Tai, C.H., et al. Evaluation of domain prediction in CASP6. Proteins 2005;61:183-192.)
    Full description of algorithm at https://ccrod.cancer.gov/confluence/display/CCRLEE/NDO]
    https://ccrod.cancer.gov/confluence/display/CCRLEE/NDO
    :param true: dict of domain assignments, with keys "linker" and "dX" where X is the domain number
    :param pred: dict of domain assignments, with keys "linker" and "dX" where X is the domain number
    example domain dictionary: {'linker':[0,1], 'D1':[2,3,4,8,9], 'D2':[5,6,7]}
    """

    # domains definitions must be mutually exclusive.
    check_unique_assignments(true)
    check_unique_assignments(pred)
    # alternative data structure would be a list of res ids for linker, plus a list of lists of res ids for domains.
    true = copy.deepcopy(true)
    pred = copy.deepcopy(pred)

    n_dom_pred = len([k for k in pred.keys() if k != "linker"])
    n_dom_gt = len([k for k in true.keys() if k != "linker"])

    # linkers are treated specially, so put them at row/col 0 to make this easy

    pred_linker = pred.pop("linker", [])
    gt_linker = true.pop("linker", [])

    gt_res_ids = [gt_linker] + list(true.values())
    pred_res_ids = [pred_linker] + list(pred.values())

    overlap = np.zeros((n_dom_pred + 1, n_dom_gt + 1))
    assert len(gt_res_ids) == overlap.shape[1] and len(pred_res_ids) == overlap.shape[0]

    for i, p_res in enumerate(pred_res_ids):
        for j, gt_res in enumerate(gt_res_ids):
            overlap[i, j] = len(set(p_res).intersection(set(gt_res)))

    # modified from v0: max overlap for a domain cannot be with a linker (hence max slices start at 1)
    if overlap.shape[0] == 1 or overlap.shape[1] == 1:
        # either the predictions or the true labels (or both) are all linker
        print('NDO score undefined for case where all linker')
        return 0
    row_scores = overlap[1:, 1:].max(axis=1)
    row_scores -= (overlap[1:].sum(axis=1) - row_scores)

    col_scores = overlap[1:, 1:].max(axis=0)
    col_scores -= (overlap[:, 1:].sum(axis=0) - col_scores)

    total_score = (row_scores.sum() + col_scores.sum()) / 2
    # count number of domain (non-linker) residues to normalize
    max_score = sum([len(gt_res) for gt_res in gt_res_ids[1:]])

    return total_score / max_score
